@@ -21,12 +21,10 @@ PosArgsT = TypeVarTuple("PosArgsT")
 
 class TaskStatus(Protocol[T_contra]):
     @overload
-    def started(self: TaskStatus[None]) -> None:
-        ...
+    def started(self: TaskStatus[None]) -> None: ...
 
     @overload
-    def started(self, value: T_contra) -> None:
-        ...
+    def started(self, value: T_contra) -> None: ...
 
     def started(self, value: T_contra | None = None) -> None:
         """
@@ -42,6 +40,12 @@ class TaskGroup(metaclass=ABCMeta):
 
     :ivar cancel_scope: the cancel scope inherited by all child tasks
     :vartype cancel_scope: CancelScope
+
+    .. note:: On asyncio, support for eager task factories is considered to be
+        **experimental**. In particular, they don't follow the usual semantics of new
+        tasks being scheduled on the next iteration of the event loop, and may thus
+        cause unexpected behavior in code that wasn't written with such semantics in
+        mind.
     """
 
     cancel_scope: CancelScope
